"""
Classes for handling data coming back from scigraph/annotations

Because the results do not include the original text the content
is added back into these objects based on the input, which is
why these are not standard dataclasses
"""

from dataclasses import dataclass, field
from typing import List


@dataclass
class Token:
    """
    A token or entity extracted from a marked-up span
    """
    id: str
    category: List[str] = field(default_factory=list)
    terms: List[str] = field(default_factory=list)


class Span:
    """
    A marked-up span of text
    """

    def __init__(self, start: int, end: int, token: Token, content=None):
        self.start = start
        self.end = end
        self.text = content[self.start:self.end]
        self.token = token
        return


@dataclass
class SciGraphAnnotation:
    token: Token
    start: int
    end: int


class EntityAnnotationResults:
    """
    The results of a SciGraph annotations/entities call
    """

    def __init__(self, results: List[SciGraphAnnotation], content=None):
        self.content = content
        self.spans = []
        for result in results:
            self.spans.append(Span(result.start, result.end, result.token, content))



import io
from ontobio.io import entitywriter, gafgpibridge, entityparser
from bin.validate import produce_gpi
from pathlib import Path
import pytest


@pytest.mark.parametrize("gpad_gpi_output_version", ["2.0", "1.2"])
def test_produce_gpi(gpad_gpi_output_version):
    # Base path relative to this script
    base_path = Path(__file__).parent / "resources"

    # Define the paths for the GAF and expected GPI file
    gaf_path = base_path / "mgi.gaf"

    # Ensure the GAF file exists to avoid FileNotFoundError
    if not gaf_path.exists():
        raise FileNotFoundError(f"Expected GAF file does not exist: {gaf_path}")

    # Set parameters for the function
    dataset = "mgi"
    ontology_graph = None  # Assuming setup elsewhere or not needed for this specific test

    # Call the function
    output_gpi_path = produce_gpi(dataset, str(base_path), str(gaf_path), ontology_graph, gpad_gpi_output_version)

    # Convert the output path to a pathlib.Path object for consistency
    output_gpi_path = Path(output_gpi_path)
    assert output_gpi_path.exists(), "The GPI file was not created."

    # Verify the contents of the GPI file
    p = entityparser.GpiParser()
    assert p.parse(open(output_gpi_path, "r")) is not None, "The GPI file could not be parsed."
    results = p.parse(open(output_gpi_path, "r"))
    assert len(results) > 5, "The GPI file should have about 9 unique genes from ~ 90 associations in the GAF file."

    with output_gpi_path.open() as f:
        lines = f.readlines()

    assert len(lines) > 0, "The GPI file should not be empty."


def test_gpi_2_0_writer():
    gpi_obj = {
        'id': "MGI:MGI:1918911",
        'label': "0610005C13Rik",  # db_object_symbol,
        'full_name': "RIKEN cDNA 0610005C13 gene",  # db_object_name,
        'synonyms': [],
        'type': ["SO:0000000"],  # db_object_type,
        'taxon': {"id": "NCBITaxon:10090"},
        'encoded_by': "", # encoded_by
        'parents': "",
        'protein_containing_complex_members': "", # protein_containing_complex_members
        'xrefs': "",
        'properties': ""
    }

    entity = gafgpibridge.Entity(gpi_obj)

    out = io.StringIO()
    gpiwriter20 = entitywriter.GpiWriter(file=out, version="2.0")
    gpiwriter20.write_entity(entity)
    outlines = out.getvalue().split("\n")

    expected_lines = [
        "!gpi-version: 2.0",
        "MGI:MGI:1918911\t0610005C13Rik\tRIKEN cDNA 0610005C13 gene\t\tSO:0000000\ttaxon:10090\t\t\t\t\t",
        ""
    ]
    assert expected_lines == outlines


def test_gpi_1_2_writer():
    gpi_obj = {
        'id': "MGI:MGI:1918911",
        'label': "0610005C13Rik",  # db_object_symbol,
        'full_name': "RIKEN cDNA 0610005C13 gene",  # db_object_name,
        'synonyms': [],
        'type': ["SO:0000000"],  # db_object_type,
        'taxon': {"id": "NCBITaxon:10090"},
        'encoded_by': "",  # encoded_by
        'parents': "",
        'protein_containing_complex_members': "",  # protein_containing_complex_members
        'xrefs': "",
        'properties': ""

    }

    entity = gafgpibridge.Entity(gpi_obj)

    out = io.StringIO()
    gpiwriter20 = entitywriter.GpiWriter(file=out, version="1.2")
    gpiwriter20.write_entity(entity)
    outlines = out.getvalue().split("\n")

    expected_lines = [
        "!gpi-version: 1.2",
        "MGI\tMGI:1918911\t0610005C13Rik\tRIKEN cDNA 0610005C13 gene\t\tSO:0000000\ttaxon:10090\t\t\t",
        ""
    ]
    assert expected_lines == outlines



"""
Utility functions for working with monarch identifiers using scigraph
"""
from typing import List, Dict, Iterator, Iterable, Optional
from json.decoder import JSONDecodeError

from ontobio.model.similarity import Node, TypedNode
from ontobio.neo.scigraph_ontology import RemoteScigraphOntology
from ontobio.util.user_agent import get_user_agent
from ontobio.model.bbop_graph import BBOPGraph
from ontobio.model.nlp import EntityAnnotationResults, SciGraphAnnotation
from ontobio.model.biomodel import BioObject, Taxon, NamedObject
from ontobio.config import get_config

from prefixcommons.curie_util import expand_uri
from dacite import from_dict
import requests
from requests import RequestException, HTTPError


# TODO: modularize into vocab/graph/etc?
HAS_PART = 'http://purl.obolibrary.org/obo/BFO_0000051'
INHERES_IN = 'http://purl.obolibrary.org/obo/RO_0000052'
INHERES_IN_PART_OF = 'http://purl.obolibrary.org/obo/RO_0002314'
HAS_ROLE = 'http://purl.obolibrary.org/obo/RO_0000087'
IN_TAXON = 'RO:0002162'
HAS_DISPOSITION = 'RO:0000091'
ENCODES = 'RO:0002205'
HAS_DBXREF = 'oboInOwl:hasDbXref'


class SciGraph:
    """
    Facade object for accessing a SciGraph instance.

    This provides access to both generic methods following a graph-oriented model, and
    domain-specific convenience methods that build in knowledge of different relationship types.

    """

    def __init__(self, url=None):
        if url is not None:
            self.url_prefix = url
        else:
            self.url_prefix = get_config()['scigraph_data']['url']

    def neighbors(self, id=None, **params):
        """
        Get neighbors of a node

        parameters are directly passed through to SciGraph: e.g. depth, relationshipType

        Returns a BBOPGraph
        """
        response = self.get_response("graph/neighbors", id, "json", **params)
        if response.status_code == 404:
            raise HTTPError

        return BBOPGraph(response.json())

    def node(self, id=None, **params):
        """
        Get a node in a graph plus its metadata

        Returns a BBOPGraph Node
        """
        response = self.get_response("graph", q=id, format="json", **params)
        if response.status_code == 404:
            raise HTTPError

        nodes = response.json()['nodes']
        if len(nodes) == 0:
            raise HTTPError

        return from_dict(NamedObject, nodes[0])

    def get_clique_leader(self, id) -> BioObject:
        """
        :raises NoResultFoundException
        :return: BioObject
        """
        response = self.get_response("dynamic/cliqueLeader", q=id, format="json")

        nodes = response.json()['nodes']
        if len(nodes) == 0:
            raise HTTPError

        return from_dict(BioObject, nodes[0])

    def bioobject(self, id, node_type=None, **params):
        """
        Get a node in a graph and translates it to biomodels datamodel

        Arguments
        ---------
        id
            identifier or CURIE

        class_name
            name of the class in the biomodel data model to instantiate

        Returns: biomodel.BioObject or subclass
        """
        bio_object = self.get_clique_leader(id)

        # get nodes connected with edge 'in_taxon'
        response = self.get_response(
            "graph/neighbors",
            q=bio_object.id,
            format="json",
            depth=1,
            relationshipType=IN_TAXON,
            direction="OUTGOING"
        )

        graph = BBOPGraph(response.json())

        bio_object.taxon = None
        for tax_edge in graph.edges:
            bio_object.taxon = from_dict(Taxon, graph.get_node(tax_edge.obj).as_dict())

        # Type specific
        if node_type == 'disease':
            # get nodes connected with edge 'has_disposition'
            response = self.get_response(
                "graph/neighbors",
                q=bio_object.id,
                format="json",
                depth=1,
                relationshipType=HAS_DISPOSITION,
                direction="OUTGOING"
            )
            graph = BBOPGraph(response.json())
            bio_object.inheritance = []
            bio_object.clinical_modifiers = []
            for disposition_edge in graph.edges:
                disposition = graph.get_node(disposition_edge.obj)
                if 'inheritance' in disposition.meta.category_list:
                    bio_object.inheritance.append(
                        from_dict(NamedObject, disposition.as_dict())
                    )
                else:
                    bio_object.clinical_modifiers.append(
                        from_dict(NamedObject, disposition.as_dict())
                    )

        return bio_object

    def graph(self, id=None):
        """
        Extracts a subgraph around a given in

        Graph includes superclass closure, equivalent classes and direct subclasses

        Returns a BBOPGraph
        """
        g1 = self.neighbors(id, relationshipType='subClassOf', blankNodes=False, direction='OUTGOING',depth=20)
        g2 = self.neighbors(id, relationshipType='subClassOf', direction='INCOMING', depth=1)
        g3 = self.neighbors(id, relationshipType='equivalentClass', depth=1)
        g1.merge(g2)
        g1.merge(g3)
        return g1

    # TODO: replace with https://github.com/SciGraph/SciGraph/issues/200
    def cbd(self, id=None):
        """
        Returns the Concise Bounded Description of a node

        See https://www.w3.org/Submission/CBD/
        """
        nodes = [id]
        g=BBOPGraph()
        while len(nodes)>0:
            n = nodes.pop()
            nextg = self.neighbors(n, params={'blankNodes':True, 'direction':'OUTGOING','depth':1})
            for nn in nextg.nodes:
                if nn.id.startswith("_:"):
                    n.append(nn.id)
            g.merge(nextg)
        return g

    def extract_subgraph(self, ids=None, relationshipType='subClassOf'):
        """
        Returns subgraph module extracted using list of node IDs as seed
        """
        ids = ids or []

        graph = BBOPGraph()
        visited = []
        while len(ids)>0:
            id = ids.pop()
            nextg = self.neighbors(id, blankNodes=False, relationshipType=relationshipType, direction='OUTGOING',depth=1)
            for edge in nextg.edges:
                next_id = edge.obj
                if next_id not in visited:
                    ids.append(next_id)
                    visited.append(next_id)
            graph.merge(nextg)
        return graph

    # TODO - direct SciGraph method?
    def traverse_chain(self, id=None, rels=None, type=None, blank=True, reverse_direction=False):
        """
        Finds all nodes reachable via a specified chain of relationship types
        """
        # we pop parts of the chain from a stack, so the
        # default direction is reversed, unless we
        # want to follow in the opposite direction

        rels = rels or []

        rels_ordered = rels.copy()
        if not reverse_direction:
            rels_ordered.reverse()

        direction = 'OUTGOING'
        if reverse_direction:
            direction = 'INCOMING'

        # list of tuples
        stack = [ (id, rels_ordered) ]

        nmap = {}
        sinks = []
        while len(stack)>0:
            (nextid, nextrels) = stack.pop()
            if len(nextrels) == 0:
                sinks.append(nextid)
            else:
                nextrel = nextrels.pop()
                nextg = self.neighbors(
                    nextid,
                    blankNodes=blank,
                    relationshipType=nextrel,
                    # works?
                    # See https://github.com/SciGraph/SciGraph/issues/135#issuecomment-305097228
                    entail=True,
                    direction=direction,
                    depth=1
                )
                for n in nextg.nodes:
                    nmap[n.id] = n
                for e in nextg.edges:
                    if not blank and e.obj.startswith(":"):
                        continue
                    stack.append( (e.obj, nextrels.copy()) )

        sinknodes = [nmap[x] for x in sinks]
        if type is not None:
            sinknodes = [x for x in sinknodes if type in x.meta.pmap['types']]

        return sinknodes


    def annotate_text(self, content, http_method='get', **args):
        """
        Directly wraps SciGraph annotations endpoint.
        Returns the text with annotated entities tagged as HTML <span>
        """

        params = {
            'content': content,
            **args
        }
        response = self.get_response("annotations", None, None, http_method, **params)
        return response.text

    def annotate_entities(self, content, http_method='get', **args):
        """
        Directly wraps SciGraph annotations/entities endpoint.
        """
        params = {
            'content': content,
            **args
        }
        response = self.get_response("annotations/entities", None, None, http_method, **params)
        scigraph_annotations = [from_dict(SciGraphAnnotation, annot) for annot in response.json()]
        return EntityAnnotationResults(scigraph_annotations, content)

    # Internal wrapper onto requests API
    def get_response(self, path="", q=None, format=None, http_method='get', **params):
        url = self.url_prefix + path
        if q is not None:
            url += "/" +q
        if format is not None:
            url = url  + "." + format
        if http_method == 'get':
            request = requests.get(url, params=params, headers={'User-Agent': get_user_agent(modules=[requests], caller_name=__name__)})
        elif http_method == 'post':
            request = requests.post(url, data=params, headers={'User-Agent': get_user_agent(modules=[requests], caller_name=__name__)})
        else:
            raise RequestException

        return request

    ## Domain-specific methods
    ## Note some of these may be redundant with https://github.com/monarch-initiative/monarch-cypher-queries/tree/master/src/main/cypher/golr-loader

    def gene_to_uniprot_proteins(self, id):
        """
        Given a gene ID, find the list of uniprot proteins that this encodes

        This method may be retired in future. See https://github.com/monarch-initiative/dipper/issues/461
        """
        uniprot_ids = []
        clique_leader = self.get_clique_leader(id)
        objs = self.traverse_chain(clique_leader.id, [ENCODES, HAS_DBXREF], blank=False, reverse_direction=False)
        for x in objs:
            if x.id.startswith('UniProtKB') and x.id not in uniprot_ids:
                uniprot_ids.append(x.id)
        return uniprot_ids

    def uniprot_protein_to_genes(self, id):
        """
        Given a UniProt ID, find the list of genes encoding this protein

        This method may be retired in future. See https://github.com/monarch-initiative/dipper/issues/461
        """
        encoding_nodes = self.neighbors(id,
                               blankNodes=False,
                               relationshipType=ENCODES,
                               # works?
                               # See https://github.com/SciGraph/SciGraph/issues/135#issuecomment-305097228
                               entail=True,
                               direction='INCOMING',
                               depth=1)
        genes = []
        for x in encoding_nodes.edges:
            genes.append(x.sub)

        # This second step is expensive and will no longer be required when
        # https://github.com/SciGraph/SciGraph/issues/135 is implemented
        gene_ids = []
        for geneId in genes:
            encoding_nodes = self.neighbors(
                geneId,
                blankNodes=False,
                relationshipType='equivalentClass',
                # See https://github.com/SciGraph/SciGraph/issues/135#issuecomment-305097228
                entail=True,
                direction='BOTH',
                depth=1
            )
            for x in encoding_nodes.edges:
                if x.sub not in gene_ids:
                    gene_ids.append(x.sub)

        return gene_ids

    def phenotype_to_entity_list(self, id):
        """
        Given a phenotype ID, find the list of affected entities

        Uses the Ontology Design Pattern has-part o inheres_in
        """
        inheres_in = self.traverse_chain(
            id,
            [HAS_PART, INHERES_IN],
            "anatomical entity"
        )
        inheres_in_part_of = self.traverse_chain(
            id,
            [HAS_PART, INHERES_IN_PART_OF],
            "anatomical entity"
        )
        objs = inheres_in + inheres_in_part_of

        return [from_dict(NamedObject, x.as_dict()) for x in objs]

    def substance_to_role_associations(self, id):
        """
        Given a chemical ID, find the list of roles

        Uses the Ontology Design Pattern CHEMICAL has-role ROLE
        """
        # TODO - include closure
        bbg = self.neighbors(id, relationshipType=HAS_ROLE, depth=1)
        return bbg_to_assocs(bbg)

    def substance_participates_in_associations(self, id):
        """
        Given a chemical ID, find the list of activities and processes that have this as participant

        Uses GO-CHEBI axioms
        """
        # TODO - include closure
        bbg = self.neighbors(id, direction='INCOMING', depth=1)
        return bbg_to_assocs(bbg)

    def get_datasets(self):
        """
        Get metadata about all the datasets in SciGraph
        """
        response = self.get_response("dynamic/datasets", None, "json")
        response_json = response.json()
        return response_json


def namespace_to_taxon() -> Dict[str, Node]:
    """
    namespace to taxon mapping
    """
    human_taxon = Node(
        id='NCBITaxon:9606',
        label='Homo sapiens'
    )
    return {
        'MGI': Node(
            id='NCBITaxon:10090',
            label='Mus musculus'
        ),
        'MONDO':   human_taxon,
        'OMIM':    human_taxon,
        'MONARCH': human_taxon,
        'HGNC':    human_taxon,
        'FlyBase': Node(
            id='NCBITaxon:7227',
            label='Drosophila melanogaster'
        ),
        'WormBase': Node(
            id='NCBITaxon:6239',
            label='Caenorhabditis elegans'
        ),
        'ZFIN': Node(
            id='NCBITaxon:7955',
            label='Danio rerio'
        )
    }


def get_scigraph_nodes(id_list)-> Iterator[Dict]:
    """
    Queries scigraph neighbors to get a list of nodes back

    We use the scigraph neighbors function because ids can be sent in batch
    which is faster than iteratively querying solr search
    or the scigraph graph/id function

    :return: json decoded result from scigraph_ontology._neighbors_graph
    :raises ValueError: If id is not in scigraph
    """
    scigraph = RemoteScigraphOntology('scigraph:data')

    chunks = [id_list[i:i + 100] for i in range(0, len(list(id_list)), 100)]
    for chunk in chunks:
        params = {
            'id': chunk,
            'depth': 0
        }

        try:
            result_graph = scigraph._neighbors_graph(**params)
            for node in result_graph['nodes']:
                yield node
        except JSONDecodeError as exception:
            # Assume json decode is due to an incorrect class ID
            # Should we handle this?
            raise ValueError(exception.doc)


def get_id_type_map(id_list: Iterable[str]) -> Dict[str, List[str]]:
    """
    Given a list of ids return their types

    :param id_list: list of ids
    :return: dictionary where the id is the key and the value is a list of types
    """
    type_map = {}
    filter_out_types = [
        'cliqueLeader',
        'Class',
        'Node',
        'Individual',
        'quality',
        'sequence feature'
    ]

    for node in get_scigraph_nodes(id_list):
        type_map[node['id']] = [typ.lower() for typ in node['meta']['types']
                                if typ not in filter_out_types]
        if not type_map[node['id']]:
            type_map[node['id']] = ['Node']

    return type_map


def get_nodes_from_ids(id_list: Iterable[str]) -> List[Node]:
    """
    Given a list of ids return their types

    :param id_list: list of ids
    :return: dictionary where the id is the key and the value is a list of types
    """
    node_list = []

    for result in get_scigraph_nodes(id_list):
        if 'lbl' in result:
            label = result['lbl']
        else:
            label = None  # Empty string or None?
        node_list.append(Node(result['id'], label))

    return node_list


def get_taxon(id: str) -> Optional[Node]:
    """
    get taxon for id

    Currently via hardcoding, should replace when scigraph when
    taxa are more universally annotated (having these as node
    properties would also be more performant)

    :param id: curie formatted id
    :return: Node where id is the NCBITaxon curie and label is the scientific name
    """
    taxon = None
    namespace = id.split(":")[0]
    if namespace in namespace_to_taxon():
        taxon = namespace_to_taxon()[namespace]

    return taxon


def typed_node_from_id(id: str) -> TypedNode:
    """
    Get typed node from id

    :param id: id as curie
    :return: TypedNode object
    """
    filter_out_types = [
        'cliqueLeader',
        'Class',
        'Node',
        'Individual',
        'quality',
        'sequence feature'
    ]
    node = next(get_scigraph_nodes([id]))

    if 'lbl' in node:
        label = node['lbl']
    else:
        label = None  # Empty string or None?

    types = [typ.lower() for typ in node['meta']['types']
             if typ not in filter_out_types]

    return TypedNode(
        id=node['id'],
        label=label,
        type=types[0],
        taxon = get_taxon(id)
    )


def bbg_to_assocs(g):
    return [bbedge_to_assoc(e,g) for e in g.edges]


def bbedge_to_assoc(e,g):
    return {
        'subject': {'id': e.sub,
                    'label': g.get_lbl(e.sub)
                    },
        'object': {'id': e.obj,
                   'label': g.get_lbl(e.obj)
                    },
        }



import pytest
import datetime
from ontobio.io import assocparser
from ontobio.io.gpadparser import to_association
from ontobio.ontol_factory import OntologyFactory
from ontobio.rdfgen.gocamgen import collapsed_assoc, gocam_builder, gocamgen

GO_ONTO = OntologyFactory().create("tests/resources/go-binding.json")  # Placeholder ontology to instantiate models
GO_ONTO.merge([OntologyFactory().create("tests/resources/ro-gp2term-20210723.json")])  # Truncated RO
PARSER_CONFIG = assocparser.AssocParserConfig(ontology=GO_ONTO,
                                              gpi_authority_path="tests/resources/mgi2.test_entities.gpi")


def test_evidence_max_date():
    ev1 = gocamgen.GoCamEvidence(code="ECO:0000314", references=["PMID:12345"], date="2008-08-14")
    ev2 = gocamgen.GoCamEvidence(code="ECO:0000314", references=["PMID:12345"], date="2011-04-12")
    ev3 = gocamgen.GoCamEvidence(code="ECO:0000314", references=["PMID:12345"], date="2021-03-01")
    max_date = gocamgen.GoCamEvidence.sort_date([ev1, ev2, ev3])[-1]
    assert max_date == "2021-03-01"


def test_get_with_froms():
    gpi_ents = gocam_builder.GoCamBuilder.parse_gpi(gpi_file="tests/resources/mgi2.test_entities.gpi")
    report = assocparser.Report(group="unknown", dataset="unknown")
    vals = [
        "MGI:MGI:1929608",
        "",
        "RO:0001025",
        "GO:0005515",
        "MGI:MGI:3028747|PMID:14627703",
        "ECO:0000353",
        "MGI:MGI:1915834",  # with/from
        "",
        "2004-04-30",
        "MGI",
        "",
        "creation-date=2004-04-30|modification-date=2004-04-30|contributor-id=http://orcid.org/0000-0002-9796-7693"
    ]
    result = to_association(list(vals), report=report, version="2.0")
    go_assoc = result.associations[0]
    ca_set = collapsed_assoc.CollapsedAssociationSet(GO_ONTO, gpi_ents)
    with_froms = ca_set.get_with_froms(go_assoc)
    assert len(with_froms) == 1 and \
           with_froms[0].header == ["MGI:MGI:1915834"] and \
           with_froms[0].line == ["MGI:MGI:1915834"]

    vals[6] = "MGI:MGI:1915834|FAKE:12345"
    result = to_association(list(vals), report=report, version="2.0")
    go_assoc = result.associations[0]
    ca_set = collapsed_assoc.CollapsedAssociationSet(GO_ONTO, gpi_ents)
    with_froms = ca_set.get_with_froms(go_assoc)
    # FAKE:12345 should be on line since not in GPI nor does it have same taxon as subject MGI:MGI:1929608
    assert len(with_froms) == 2 and \
           with_froms[0].header == ["MGI:MGI:1915834"] and \
           with_froms[0].line == ["MGI:MGI:1915834"] and \
           with_froms[1].header == [] and \
           with_froms[1].line == ["FAKE:12345"]

    # Test merging of same-header with/from values in different order
    ca_set = collapsed_assoc.CollapsedAssociationSet(GO_ONTO, gpi_ents)
    header1 = collapsed_assoc.GoAssocWithFrom(header=["MGI:MGI:1915834", "FAKE:12345"])
    header2 = collapsed_assoc.GoAssocWithFrom(header=["FAKE:12345", "MGI:MGI:1915834"])
    ca_set.find_or_create_collapsed_association(go_assoc, with_from=header1)
    ca_set.find_or_create_collapsed_association(go_assoc, with_from=header2)
    assert len(ca_set.collapsed_associations) == 1 and ca_set.collapsed_associations[0].with_froms == ["FAKE:12345", "MGI:MGI:1915834"]


def test_ref_picker():
    test_refs = [
        "GO_REF:0000483",
        "doi:485930",
        "WB_REF:WBPaper00003384",
        "PMID:9834189",
    ]
    result = gocamgen.ReferencePreference.pick(test_refs)
    assert result == "PMID:9834189"

    test_refs = [
        "GO_REF:0000483",
        "doi:485930",
        "PMID:9834189",
        "WB_REF:WBPaper00003384",
    ]
    result = gocamgen.ReferencePreference.pick(test_refs)
    assert result == "PMID:9834189"

    test_refs = ["ZFIN:ZDB-PUB-170709-3"]
    result = gocamgen.ReferencePreference.pick(test_refs)
    assert result == "ZFIN:ZDB-PUB-170709-3"


def test_model_title():
    builder = gocam_builder.GoCamBuilder(parser_config=PARSER_CONFIG, modelstate="test")
    title = builder.model_title(gene_id="MGI:MGI:1915834")
    assert title == "1110020C17Rik (MGI:MGI:1915834)"

    # Fallback to gene_id as title if not found in GPI
    title = builder.model_title(gene_id="FAKE:1915834")
    assert title == "FAKE:1915834"


def test_model_dates():
    model_associations = []
    version = "2.0"
    report = assocparser.Report(group="unknown", dataset="unknown")
    vals = [
        "MGI:MGI:1915834",
        "",
        "RO:0002327",
        "GO:0003674",
        "MGI:MGI:2156816|GO_REF:0000015",
        "ECO:0000307",
        "",
        "",
        "2020-10-09",
        "MGI",
        "",
        "creation-date=2020-09-17|modification-date=2020-10-09|contributor-id=http://orcid.org/0000-0003-2689-5511"
    ]
    model_associations = model_associations + to_association(list(vals), report=report, version=version).associations

    # Different term, earlier creation-date
    vals[3], vals[11] = "GO:0016301", "creation-date=2011-12-13"
    model_associations = model_associations + to_association(list(vals), report=report, version=version).associations

    # Different term, no annotation properties
    vals[3], vals[11] = "GO:0001962", ""
    model_associations = model_associations + to_association(list(vals), report=report, version=version).associations

    builder = gocam_builder.GoCamBuilder(parser_config=PARSER_CONFIG, modelstate="test")
    model = builder.translate_to_model(gene="MGI:MGI:1915834", assocs=model_associations)
    assert model.date == "2020-10-09"
    assert model.creation_date == "2011-12-13"
    assert model.import_date == datetime.date.today().isoformat()



"""
Classes for exporting entities.

So far only one implementation
"""
import re

def stringify(s):
    if s is None:
        return ""
    elif isinstance(s,list):
        return "|".join(s)
    else:
        return s

external_taxon = re.compile("taxon:([0-9]+)")
internal_taxon = re.compile("NCBITaxon:([0-9]+)")

def normalize_taxon(taxon):
    global internal_taxon
    global external_taxon

    if external_taxon.match(taxon):
        # If we match here, then the internal view already exists and we're good
        return internal_taxon

    match = internal_taxon.match(taxon)
    if match:
        taxon_id = match.group(1)
        return "taxon:{num}".format(num=taxon_id)

    return taxon


class EntityWriter():
    """
    Abstract superclass of all association writer objects (Gpad, GAF)
    """

    # TODO: add to superclass
    def _split_prefix(self, ref):
        id = ref['id']
        [prefix, local_id] = id.split(':', maxsplit=1)
        return prefix, local_id

    # TODO: add to superclass
    def _write_row(self, vals):
        vals = [stringify(v) for v in vals]
        line = "\t".join(vals)
        self.file.write(line + "\n")

    # TODO: add to superclass
    def write_entity(self, e):
        """
        Write a single entity
        """
        pass  ## Implemented in subclasses

    def write(self, entities, meta=None):
        """
        Write a complete set of entities to a file

        Arguments
        ---------
        entities: list[dict]
            A list of entity dict objects
        meta: Meta
            metadata about association set (not yet implemented)

        """
        for e in entities:
            self.write_entity(e)


class GpiWriter(EntityWriter):
    """
    Writes entities in GPI 1.2 or 2.0 (https://github.com/geneontology/go-annotation/blob/master/specs/gpad-gpi-2-0.md) format

    :param file: file
    :param version: str

    Takes an "entity" dictionary generated typically from a GoAssociation object

    {
        'id': id, (String)
        'label': db_object_symbol, (String)
        'full_name': db_object_name, (String)
        'synonyms': synonyms, (List[str])
        'type': db_object_type, (String)
        'parents': parents, (List[Str])
        'xrefs': xref_ids, (List[Str])
        'taxon': {
            'id': self._taxon_id(taxon) (String)
        }
    }
    """
    def __init__(self, file=None, version=None):
        self.file = file
        self.version = version
        if self.file:
            if self.version == "2.0":
                self.file.write("!gpi-version: 2.0\n")
            else:
                self.file.write("!gpi-version: 1.2\n")

    def write_entity(self, entity):
        """
        Write a single entity to a line in the output file

        :param entity: dict ; typically a dictionary representing an instance of a GoAssociation object
        :param gpi_output_version: str ; the version of the GPAD output file to write
        :return: None

               GPI 2.0 spec <-- entity attributes
               
            1. DB_Object_ID <-- entity.id (CURIE format)
            2. DB_Object_symbol <-- entity.label
            3. DB_Object_Name <-- entity.full_name
            4. DB_Object_Synonyms <-- entity.synonyms
            5. DB_Object_Type <-- entity.type
            6. DB_Object_Taxon <-- entity.taxon
            7. Encoded_by <-- does not appear in GAF file, this is optional in GPI
            8. Parent_Protein <-- entity.parents # unclear if this is a list or a single value
            9. Protein_Containing_Complex_Members <-- does not appear in GAF file, this is optional in GPI
            10. DB_Xrefs <-- entity.xrefs
            11. Gene_Product_Properties <-- entity.properties

                GPI 1.2 spec <-- entity attributes

            1. DB <-- entity.id.prefix
            2. DB_Object_ID	 <-- entity.id.local_id
            3. DB_Object_Symbol <-- entity.label
            4. DB_Object_Name <-- entity.full_name
            5. DB_Object_Synonym(s) <-- entity.synonyms
            6. DB_Object_Type <-- entity.type
            7. Taxon <-- entity.taxon
            8. Parent_Object_ID <-- entity.parents # unclear if this is a list or a single value
            9. DB_Xref(s) <-- entity.xrefs
            10. Properties <-- entity.properties

        """

        if self.version == "2.0":
            vals = [
                entity.get('id'),  # DB_Object_ID
                entity.get('label'),  # DB_Object_symbol
                entity.get('full_name'),  # DB_Object_Name
                entity.get('synonyms'),  # DB_Object_Synonyms
                entity.get('type'),  # DB_Object_Type
                normalize_taxon(entity.get("taxon").get("id")),  # DB_Object_Taxon
                "",  # Encoded_by
                entity.get('parents'),  # Parent_Protein
                "",  # Protein_Containing_Complex_Members
                entity.get('xrefs'),  # DB_Xrefs
                entity.get('properties')  # Gene_Product_Properties
            ]
        else:
            prefix, local_id = self._split_prefix(entity)
            vals = [
                prefix,  # DB
                local_id,  # DB_Object_ID
                entity.get('label'),  # DB_Object_Symbol
                entity.get('full_name'),  # DB_Object_Symbol
                entity.get('synonyms'),  # DB_Object_Name
                entity.get('type'),  # DB_Object_Synonyms
                normalize_taxon(entity.get("taxon").get("id")),  # taxon
                entity.get('parents'),  # Parent_Object_ID
                entity.get('xrefs'),  # DB_Xref(s)
                entity.get('properties')  # Properties
            ]

        self._write_row(vals)



"""
Dataclasses used in scigraph_util

These were originally autogenerated from the biolink-api
marshmallow definitions

See original classes here:
https://raw.githubusercontent.com/biolink/biolink-api/1e81503/biomodel/core.py
"""

from dataclasses import dataclass, field
from typing import List, Dict, Optional

from ontobio.util.curie_map import get_curie_map

from prefixcommons.curie_util import expand_uri


@dataclass
class NamedObject:
    """
    NamedObject

    Arguments
    ---------
     id :

         ID or CURIE e.g. MGI:1201606

     label :

         RDFS Label

     iri :

         IRI

     category :

         Type of object

     description :

         Description (or definition) of an object

     types :

         Type of object (direct)

     synonyms :

         list of synonyms or alternate labels

     replaced_by :

         Direct 1:1 replacement (if named object is obsoleted)

     consider :

         Potential replacement object (if named object is obsoleted)
    """
    id: str = None
    lbl: Optional[str] = None
    iri: Optional[str] = None
    category: Optional[str] = None
    description: Optional[str] = None
    types: Optional[List[str]] = field(default_factory=list)
    synonyms: Optional[List[str]] = field(default_factory=list)
    deprecated: Optional[bool] = None
    replaced_by: Optional[List[str]] = field(default_factory=list)
    consider: Optional[List[str]] = field(default_factory=list)
    meta: Optional[Dict] = field(default_factory=dict)

    def __post_init__(self):
        """
        Logic moved from scigraph_util.map_tuple
        """
        if self.lbl:
            self.label = self.lbl

        if 'category' in self.meta:
            self.category = self.category or self.meta['category']

        self.iri = self.iri or expand_uri(self.id, [get_curie_map()])

        if 'synonym' in self.meta:
            self.synonyms = self.synonyms or [
                SynonymPropertyValue(pred='synonym', val=s) for s in self.meta['synonym']
            ]

        if 'definition' in self.meta and self.meta['definition']:
            self.description = self.description or self.meta['definition'][0]

@dataclass
class Taxon:
    id: Optional[str] = None
    lbl: Optional[str] = None

    def __post_init__(self):
        if self.lbl:
            self.label = self.lbl



@dataclass
class BioObject(NamedObject):
    """
    BioObject

    Superclass: NamedObject
    """

    taxon: Optional[Taxon] = None
    association_counts: Optional[int] = None
    xrefs: Optional[List[str]] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()

        if 'http://www.geneontology.org/formats/oboInOwl#hasDbXref' in self.meta:
            self.xrefs = self.xrefs or self.meta['http://www.geneontology.org/formats/oboInOwl#hasDbXref']


@dataclass
class Disease(BioObject):
    inheritance: Optional[List[NamedObject]] = field(default_factory=list)
    clinical_modifiers: Optional[List[NamedObject]] = field(default_factory=list)


@dataclass
class AbstractPropertyValue:
    """
    AbstractPropertyValue

    Arguments
    ---------
     val :

         value part

     pred :

         predicate (attribute) part

     xrefs :

         Xrefs provenance for property-value
    """

    val: Optional[str] = None
    pred: Optional[str] = None
    xrefs: Optional[List[str]] = field(default_factory=list)


@dataclass
class SynonymPropertyValue(AbstractPropertyValue):
    """
    SynonymPropertyValue

    Superclass: AbstractPropertyValue
    """
    id: Optional[str] = None



"""
MME response object, see
https://github.com/ga4gh/mme-apis/blob/87f615/search-api.md
"""

from dataclasses import dataclass, field
from typing import List

from ontobio.model.mme.request import Patient


@dataclass
class Score:
    patient: int


@dataclass
class Result:
    score: Score
    patient: Patient = None
    _disease: str = None
    _disease_label: str = None
    _gene: str = None
    _gene_label: str = None

@dataclass
class MmeResponse:
    disclaimer: str = None
    terms: str = "https://monarchinitiative.org/about/disclaimer"
    results: List[Result] = field(default_factory=list)
